/*
 *	Videot Encoder
 *
 *	The basic design is straight out of Mike Singletons 
 *	"Doomdarks Revenge" from back in the early 1980's. Mike used this
 *	scheme to encode the 3d landscape objects to fit in a 48K spectrum.
 *	Other folks used the same idea for faces soon after.
 *
 *	We add the obvious xor option for progressive changes (may not be a win)
 *	and implicitly thus shorter encoding for unchanged lines. Adding encodings for
 *	slide left, slide right, and bigger/smaller is the next project if
 * 	needed. Firstly I want to see if zlib can naturally encode stuff for us.
 *
 *	Another interesting option is "previous line scaled" as faces tend
 *	to smoothly flow out and back in, that may let us turn over half the
 *	image into 16 bit slots. If we go to bitpacking maybe down to 12bits
 *
 *	Should we use constant bitfiltering for specks or should we use a 
 *	curve so that we ignore more on the edges ?
 *	-	Trying curves for now.
 *
 *	BEND mode is not supported in the current decoder, and it doesnt work in the
 *	encoder either!
 */


/*
 *	ENCODER options
 */
 
#define BROADBAND		/* 9600 baud or higher */


#ifdef BROADBAND	/* Bwhwhahaha */

/*
 *	High bandwidth mode... (about 1300 bytes/frame after compression)
 */
 
#define DITHER		/* Go for filtering not dithering of borderline bits */
#define MAX_CUT	16	/* Largest vector array allowed per line */
#define BITCLAMP 2
#elif MIDDLING

/*
 *	This pair gives you about 1000-1300bytes/frame dropping to 800 when compressed
 *	Going to MAX_CUT 8 and no dither materially drops the bandwidth but makes
 *	the codec extremely face-centric, and you'll want a solid light background
 *	illuminated from in front of the speaker.
 */

#define DITHER
#define MAX_CUT 12
#define BITCLAMP 2
#else

/*
 *	You must have clear backgrounds for this, but compressed it takes the video
 *	down to 550 bytes a frame or less.
 */
#define MAX_CUT 8
#define BITCLAMP 4
#endif


#include <stdio.h> 
#include "vidiot.h"

#define min(a,b)	(((a)>(b))?(b):(a))

static int pads;
static int padms;
static int bends;
static int xors;
static int lines;

static void emit_ypad(struct video *v, int yold)
{
	while(yold>=4)
	{
		int n=yold;
		if(n>8)
			n=8;
		if(n&1)
			n--;
		outbyte(n/2-1);
		yold-=n;
		padms++;
	}
	while(yold)
	{
		outbyte(0);
		outbyte(v->width);
		yold--;
		pads++;
	}
}

static int count_span(u8 *data, int len, int polarity)
{
	int count=0;
	while(*data++==polarity && count < len)
		count++;
	if(count==0)
	{
		fprintf(stderr, "Polarity botch\n");
		exit(1);
	}
	return count;
}
		
static int encode_data_bits_plain(u8 *out, u8 *data, int w, int cut)
{
	int x=0;
	int tspan=0;
	int polarity=0, oldpolarity=0;
	int count=0;

	while(x<w)
	{
		int s;
		int high=1;
	
		if(x+w < w / 4)
			high=1; /* Boring data */
		else if(x > (3*w)/4)
			high=1;
		if(x+w < w / 3)
			high=2; /* Less boring data */
		else if(x > (2*w)/3)
			high=5;
		else
			high=6;
	
		s=count_span(data+x, w-x, polarity);
		if(s<(cut*3)/high || polarity == oldpolarity)
			tspan+=s;
		else
		{
			out[count++]=tspan;
			tspan=s;
			oldpolarity=polarity;
		}
		x+=s;
		polarity^=1;
	}
	if(tspan)
		out[count++]=tspan;
	return count;
}
		

static int encode_data_plain(struct video *v, int y, u8 *buf, int *bsv)
{
	int bslice=1;
	int pass=0;
	int count=0;

	do
	{
		count = encode_data_bits_plain(buf, v->row[y], v->width, bslice);
		bslice++;
	}
	while(count>MAX_CUT);

	/*
 	 *	We have discarded detail to get down to 8 bytes or less per
 	 *	row
 	 */
 	
 	*bsv = bslice;
 	return count;
}
 
static int encode_data_xor(struct video *v, int y, u8 *buf, int cut)
{
	int count;
	char block[256];
	int i;

	for(i=0;i<v->width;i++)
	{
		if(v->row[y][i] ^ v->oldrow[y][i])
			block[i]=1;
		else
			block[i]=0;
	}
	count = encode_data_bits_plain(buf, block, v->width, cut);

 	return count;
}

/*
 *	Pixmagic processing
 */
 
static int pixmagic_compute_line(u8 *in, int w)
{
	int pm = 0;
	int x = 0;
	int len = w;
	int ct = 0;

	while(ct++ < len)
	{
		pm += *in++;
		x++;
	}
	return pm/len;
}

static int pixmagic = 0;
static int pixhistory = 0;

/*
 *	Find the pixmagic for a line
 */
 
void compute_pixmagic(struct video *v)
{
	int y = 0;
	int pm = 0;;

	while(y< v->height)
	{
		pm+=pixmagic_compute_line(v->row[y], v->width);
		y++;
	}

	pm/=v->height;

	if(pixhistory && abs(pm-pixmagic) < 32)
		pixmagic = (3*pixmagic/4)+pm/4;
	else
		pixmagic = pm;
	pixhistory = 1;

	fprintf(stderr, "Pixmagic is %d.\n", pixmagic);
}


/*
 *	Process a line for its pixmagic
 */
 
static void pixmagic_line(struct video *v, int y)
{
	int x = 0;
	int lpm = 0;

	while(x < v->width)
	{
#ifdef DITHER 
		if(abs(v->row[y][x] - pixmagic) < 15)
		{
			/* You are as they say.. not expected to understand how this works.. */
	
			if(y && v->row[y][x] < pixmagic+15 && v->oldrow[y][x]==0 && v->row[y-1][x]==1)
			{
				v->row[y][x] = 0;
				if((x&1) && (y&1))
					fprintf(stderr, " ");
				x++;
				continue;
			}
			if(y && v->row[y][x] > pixmagic-15 && v->oldrow[y][x]==1 && v->row[y-1][x]==1)
			{
				v->row[y][x] = 1;
				if((x&1) && (y&1))
					fprintf(stderr, "X");
				x++;
				continue;
			}
		}
#endif	
		if(v->row[y][x] < pixmagic)
		{
			if((x&1) && (y&1))
			{
				fprintf(stderr, "X");
			}
			v->row[y][x]=0;
			lpm = 0;
		}
		else
		{
			if((x&1) && (y&1))
			{
				fprintf(stderr, " ");
			}
			v->row[y][x]=1;
			lpm = 1;
		}
		x++;
	}
	if(y&1)
		fprintf(stderr, "\n");
	if(v->row[y][0])
		fprintf(stderr, "Edge botch.\n");
}
	
static int bend_line(struct video *v, int y, u8 *buf, int l, int r, u8 *oldcode, int oldcost, int bslice)
{
	int oldl = oldcode[0];
	int oldr = oldcode[oldcost-1];
	int ow = v->width - (oldr + oldl);
	int w = v->width - (l + r);
	int c = 0;
	char tmp[256], decode[256];
	int slew = l - oldl;
	int bend;
	int oldx = 0;
	int newx;
	int lnx = 0;
	int x = 0;
	int dc = 0;
	int polarity = 0;

	/* Nothing to encode */
	if(ow ==0 || oldcost == 1)
		return 10000;
	
	bend = (w*256)/ow;

	/* We only have 4 bits of slew */
	if ( abs(slew)> 7)
		return 10000;
	
	/* And 6 bits of bend */
	if ( abs(256-w) > 31)
		return 10000;
	/*
	 *	Bend the line
	 */

	while(c < oldcost-1)
	{
		lnx = newx;
		oldx += oldcode[c];
		newx = (oldx * bend) / 256;
		buf[c] = newx-oldx;
		c++;
	}
	buf[c] = v->width - newx;	/* Make it add up */

	/*
	 *	Decode it
	 */ 

	while(x < v->width)
	{
		int i;
	
		for(i=0; i < buf[dc]; i++)
		{
			decode[x+i] = polarity ^ v->row[y][x+i];
		}
		polarity^=1;
		x+=buf[dc];
	}

	/*
 	 *	We now have an error map.
 	 */
	
	x=0;
	polarity = 0;

	while(x, v->width)
	{
		dc = count_span(decode+x, v->width - x, bslice);
		x += dc;
		if(dc > bslice && polarity==1)
			return 10000;	/* Don't encode this way - over error */
		polarity^=1;
	}

	bend = 256 - bend;		/* 256 is 1:1 */

	buf[0] = 4 | ((slew<0)?2:0) | (bend < 0);
	buf[1] = abs(slew)<<5 |abs(bend);
	return 2;
}

/*
 *	Encode a single frame. The input frame is arbitary height and width,
 *	The width must be below 256 pixels. The left 8 pixels of the image
 *	must be black and are not displayed. Typically you want to send
 *	240x320 images so encode 248 wide. The maximum cost at 240x320 is
 *	2.5K per frame. The typical cost is well below that which seems nice
 *	for IPAQ full screen video conferencing. For postage stamp sizes
 *	of say 120x160 the bitcost comes down to 1.2K per frame worst case.
 */
 
void videot_queue_frame(struct video *v)
{
	int y = 0;
	u8 buf1[256], buf2[256], buf3[256];
	int yold = 0;
	int oldcost = -1 ;
	u8 oldcode[256];
	int bitslice;

	/* Firstly copy the data from current to old summing a pixmagic
	   as we go, then we factor that into the old pixmagic */

	fprintf(stderr, "\033[0;0H\033[JBegin frame(%d x%d) .\n",
		v->width, v->height);

	compute_pixmagic(v);

	/*
	 *	Now encode the image.
	 */

	while(y<v->height)
	{
		int cost1, cost2;
		int bend = 10000;
	
		pixmagic_line(v, y);
		cost1 = encode_data_plain(v, y, buf1, &bitslice);
		cost2 = encode_data_xor(v, y, buf2, min(bitslice, BITCLAMP))+1;
	
		if(oldcost != -1 && 0)
			bend = bend_line(v, y, buf3, buf1[0], buf1[cost1-1], oldcode, oldcost, bitslice);
	
		if(cost2 == 2)	/* This also means 'no change' */
		{
			yold++;
			y++;
			continue;
		}

#ifdef SYNC_TEST
		outbyte(255);		
		outbyte(254);		
#endif
		emit_ypad(v, yold);
		yold = 0;
	
		if(bend < cost1 && bend < cost2)
		{
			outbuffer(buf3, 2);
			bends++;
		}
		else if(cost1 < cost2)	
		{
			oldcost = cost1;
			memcpy(oldcode, buf1, cost1);
			outbuffer(buf1, cost1);
			lines++;
		}
		else
		{
			oldcost = -1;
			outbyte(0);
			outbuffer(buf2, cost2-1);
			xors++;
		}
		y++;
#ifdef SYNC_TEST	
		outbyte(255);		
		outbyte(255);		
#endif	
	}
	emit_ypad(v, yold);

	fflush(stdout);
	fprintf(stderr, "End frame:");
	outeof();
	fprintf(stderr, "Bends %d Xors %d Lines %d.\n", bends, xors, lines);
	sleep(1);

}
